# toyMS.R
#
# Consult 2011 file, toymix.R for the earlier mixture model version.
# This version produces the Markov-switching version of the results.
#
# Successor to toymix.R -- using an MS model to illustrate forecast
# comparisons and DGPs
#
# Patrick T. Brandt
#
#
# 20120228 : Initial version (based on toymix.R from POLMETH 2011)
# 20120613 : Updated to work with latest MSBVAR devel and to
#            empirically compute the PITs
# 20120719 : Run another analysis for the MSBVAR updates.
# 20130920 : Corrected comments on the DGP
# 20130921 : Add more comments what the script does.
#

# Load libraries
library(MSBVAR)

# Set sample size
TT <- 550

# Set parameters for switching
m <- 2  # number of endogenous variables
p <- 1  # lag length
h <- 2  # number of regimes

# Setup transition matrix with two states

Q <- matrix(c(.8, .2,
              .05, .95), nrow=h, byrow=TRUE)

# Simulate MS TS data

# N(-1, 0.5) v. N(1, 0.5)...  average out to N(0, 0.5)... but are
# distinct, but made very simple.

######################################################
# MSVAR Sim
######################################################
# theta stores paramater values
# 1st column is intercept
# 2:m*p are the AR coefficients
# (mp+2)'th columns are variance

# regime 1
var.beta0.st1  <- c(-1,1)  # intercepts
var.betas.st1  <- matrix(c(.7, .1,
                           .7, .1), m, byrow=TRUE)
# regime 2
var.beta0.st2  <- c(1,-1)  # intercepts
var.betas.st2  <- matrix(c(.2, .1,
                           .2, .1), m, byrow=TRUE)

# Build the array
var.beta0 <- array(NA, c(m,1,h))
var.betas <- array(NA, c(m,p*m,h))
var.beta0[,,1] <- var.beta0.st1
var.beta0[,,2] <- var.beta0.st2
var.betas[,,1] <- var.betas.st1
var.betas[,,2] <- var.betas.st2

# Variance-Covariance Matrix for
# MVN distributed disturbances
# regime 1
e.vcv.st1  <- 0.5*diag(m)

# regime 2
e.vcv.st2  <- 0.5*diag(m)

# combine variances
e.vcv <- array(NA, c(m, m, h))
e.vcv[,,1] <- e.vcv.st1
e.vcv[,,2] <- e.vcv.st2

# hold true values of parameters for easy comparison to estimates
theta.true.var <- array(NA, c(m, 1+m*p+m, h))
theta.true.var[,1,] <- var.beta0
theta.true.var[,2:(1+p*m),] <- var.betas
theta.true.var[,(1+m*p+1):ncol(theta.true.var),] <- e.vcv

# Generate the simulated MSVAR data
set.seed(81571)
simdata <- simulateMSVAR(bigt=TT, m, p, var.beta0, var.betas, e.vcv, Q)

# Set training set and ex ante forecast horixon
Y <- ts(simdata[[1]])
Y.past <- window(Y, start=1, end=500)
Y.future <- window(Y, start=501, end=520)

# Plot data
pdf(file="toyMSdata.pdf", width=8, height=6)
par(mfrow=c(2,1))
plot(window(Y, end=c(520)), plot.type="s", lty=1:2, main="Data Series",
     ylab="", xlab="")
abline(h=0)
abline(h=3, lty=2)
abline(h=-3, lty=2)
abline(v=500, lty=2)
plot(window(ts(simdata[[2]]-1), end=c(520)),
            main="Regime Probabilities",
            ylab="Pr(Regime=1)", xlab="")
abline(v=500, lty=2)
dev.off()

################################################
# Estimate models, based on the training set
################################################

# Estimate posterior mode under a diffuse prior
set.seed(1234)
M0 <- msbvar(Y.past, p=1, h=2, lambda0=1, lambda1=1, lambda3=1,
             lambda4=1, lambda5=1, mu5=0, mu6=0, qm=4,
             alpha.prior=matrix(c(10,1,1,10), 2, 2), prior=2,
             max.iter=50)


# Now plot the EM v. actual regime probs
plot(ts(cbind(M0$fp,(simdata[[2]]-1)[2:500])),
     plot.type="s", col=1:3, ylab="Regime probailities and truth")

# Gibbs sampler the posterior
N1 <- 1000
N2 <- 1000

# Now fit the posterior density.
set.seed(231786)
M0.gibbs <- gibbs.msbvar(M0, N1, N2, permute=TRUE)

plotregimeid(M0.gibbs, type="intercepts")

# Now sample with identification
M0id.gibbs <- gibbs.msbvar(M0, N1, N2,
                           permute=FALSE, Beta.idx=c(3,1))


# Get summaries
summary(M0id.gibbs$Q.sample)
summary(M0id.gibbs$Beta.sample)
summary(M0id.gibbs$Sigma.sample)

par(mfrow=c(2,1))
plot.SS(M0id.gibbs)
plot(ts(simdata[[2]][1:500]-1), ylab="True Regimes")

###################################################
# Forecast
#
# 1) DGP forecast based on MS model
# 2) Normal linear projection (no switching) from a VAR
# 3) Incorrect N(0,1)

#source("~/MSBVARlocal/devel-unstable/msbvarforecast.R")
N1 <- 1000
N2 <- 5000
k <- 20
system.time(ftmp <- forecast(M0id.gibbs, k, N1=N1,
                             N2=N2))

# Now summaries of forecasts
fout <- apply(ftmp$forecast, c(2,3), quantile, probs=c(0.05, 0.5, 0.95))

rd <- Y.future[1:k,]

par(mfrow=c(2,1))
plot(ts(cbind(rd[,1],t(fout[,,1]))), plot.type="s", col=c(1,2,2,2),
     lty=c(1,2,1,2), ylab="Series 1")
plot(ts(cbind(rd[,2],t(fout[,,2]))), plot.type="s", col=c(1,2,2,2),
     lty=c(1,2,1,2), ylab="Series 2")


#cbind(mean.SS(ftmp), (simdata[[2]]-1)[2:500+k])

plot(window(ts(cbind(mean.SS(ftmp), (simdata[[2]]-1)[2:(500+k)])),
            start=450),
     plot.type="s", col=1:(h+1),
            main="Plot of last 50 data points and 20 forecasts",
            ylab="Regime probabilities")
abline(v=500, lty=2)


################################################################
# Other forecasts, based on the sample
#
# 1) Normal draws, mixed
# 2) VAR (no switching)
# 3) Sample means & variances
################################################################

# Find steady state for mixture in MS process
steady.Q <- function(P)
  { M <- dim(P)[1]
    if (M<3)
      {
        eta <- solve(rbind(cbind(1 - P[1,1], P[2,1]),
                           rep(1,2)))%*%matrix(c(0,1))
      }
    else
      {
        eta <- solve(rbind(cbind(diag(M-1) - t(P)[1:(M-1),1:(M-1)],
                                 t(P)[1:(M-1),M]),
                           rep(1,M)))%*%matrix(c(rep(0,M-1),1))
      }
    # Find the steady state if there are negative values -- that is
    # iterate a bit!
    while(cumprod(eta)[M]<0)
      {
        eta <- t(P)%*%eta
      }
    return(eta)
  }

mix <- steady.Q(Q)

forc.norm1 <- array(rmvnorm(N2*k, mean=c(-1,1), sigma=0.5*diag(2)),
                    c(N2, k, 2))

forc.norm2 <- array(rmvnorm(N2*k, mean=c(1,-1), sigma=0.5*diag(2)),
                    c(N2, k, 2))

# Normal mixture, no MS
forc.norm <- forc.norm1*mix[1] + forc.norm2*mix[2]


# VAR forecast
VAR1 <- szbvar(Y.past, p=1, lambda0=1, lambda1=0.1, lambda3=1,
               lambda4=0.25, lambda5=0, mu5=0,
               mu6=0, qm=4, prior=0)

fVAR <- forecast(VAR1, nsteps=k)

fVARuc <- uc.forecast(VAR1, nsteps=k, burnin=N1, gibbs=N2)
fVAR <- fVARuc$forecast

#############################################################
# Incorrect sample mean forecasts
#############################################################
mn <- apply(Y.past, 2, mean)
covs <- cov(Y.past)

tmp <- rmvnorm(m*k*N2, mean=mn, sigma=covs)
normonly <- array(0, c(N2, k, m))
normonly[,,1] <- array(tmp[,1], c(N2, k, 1))
normonly[,,2] <- array(tmp[,2], c(N2, k, 1))

#############################################################
# Compute the summary stats over the sample of forecasts:
# 1) Average RMSE
# 2) Average MAE
# 3) Average Gaussian approx.
# 4) Empirical CDF
#############################################################
TT <- nrow(Y.past)
MS <- ftmp$forecast
VAR <- fVAR

#############################################################
# Average RMSEs / MAE / over k forecasts
#############################################################
rmse.VAR <- rmse.norm <- rmse.MS <- matrix(0, 1, 2)
mae.VAR <- mae.norm <- mae.MS <- matrix(0, 1, 2)
for(i in 1:m)
{
    # RMSE
    rmse.MS[1,i] <- mean(sqrt(apply((MS[,,i] -
                                     as.vector(Y.future[,i]))^2, 1,
                                    mean)))
    rmse.VAR[1,i] <- mean(sqrt(apply((VAR[,,i] -
                                     as.vector(Y.future[,i]))^2, 1,
                                    mean)))

    rmse.norm[1,i] <- mean(sqrt(apply((normonly[,,i] -
                                       as.vector(Y.future[,i]))^2, 1,
                                      mean)))

    # MAE
    mae.MS[1,i] <- mean(abs(MS[,,i] - as.vector(Y.future[,i])))
    mae.VAR[1,i] <- mean(abs(VAR[,,i] - as.vector(Y.future[,i])))
    mae.norm[1,i] <- mean(abs(normonly[,,i] - as.vector(Y.future[,i])))
}

#####################################################
# CRPS
#####################################################
library(verification)

# Compute the CRPS for each forecast set and variable
MS1.crps <- crps(as.vector(Y.future[,1]),
                 cbind(apply(MS[,,1], 2, mean),
                       apply(MS[,,1], 2, sd)))

MS2.crps <- crps(Y.future[,2],
                 cbind(apply(MS[,,2], 2, mean),
                       apply(MS[,,2], 2, sd)))

VAR1.crps <- crps(Y.future[,1],
                  cbind(apply(VAR[,,1], 2, mean),
                        apply(VAR[,,1], 2, sd)))

VAR2.crps <- crps(Y.future[,2],
                  cbind(apply(VAR[,,2], 2, mean),
                        apply(VAR[,,2], 2, sd)))


norm1.crps <- crps(Y.future[,1],
                   cbind(apply(normonly[,,1], 2, mean),
                         apply(normonly[,,1], 2, sd)))

norm2.crps <- crps(Y.future[,2],
                   cbind(apply(normonly[,,1], 2, mean),
                         apply(normonly[,,1], 2, sd)))



# Compute / plot the PITs

# First, used ecdfs to compute the PIT values and then plot.

# Storage for values
MS.pit <- VAR.pit <- norm.pit <- matrix(0, k, 2)

# Do this by variable across the time periods
for(i in 1:m)
{
    tmp <- matrix(as.vector(Y.future[,i]), byrow=TRUE, nrow=N2, ncol=k)
    MS.pit[,i] <- apply(MS[,,i] < tmp, 2, mean)
    VAR.pit[,i] <- apply(VAR[,,i] < tmp, 2, mean)
    norm.pit[,i] <- apply(normonly[,,i] < tmp, 2, mean)
}

# Second, check the PITs for autocorrelation
par(mfrow=c(3,2))
apply(MS.pit, 2, acf)
apply(VAR.pit, 2, acf)
apply(norm.pit, 2, acf)

# Compute the KS tests on the PITS
MS.ks <- apply(MS.pit, 2, ks.test, y="punif")
VAR.ks <- apply(VAR.pit, 2, ks.test, y="punif")
norm.ks <- apply(norm.pit, 2, ks.test, y="punif")

# Third, plot the PITS
pdf(file="toyMSPIT.pdf")
vnames <- c("Series 1", "Series 2")

par(mfcol=c(3,2), mai=c(0.4,0.25,0.25,0.25), omi=c(0,0.75,0,0))
for(i in 1:m)
{
    hist(MS.pit[,i], main=vnames[i], breaks=seq(0,1, by=0.05),
         prob=TRUE, xlab="", ylab="")
    if(i==1) mtext("MSBVAR", side=2, line=3)
    mtext(paste("KS, p = ", round(MS.ks[[i]]$p.value,3)),
          side=1, line=2, cex=0.7)

    hist(VAR.pit[,i], main="", breaks=seq(0,1, by=0.05),
         prob=TRUE, xlab="", ylab="")
    if(i==1) mtext("VAR", side=2, line=3)
    mtext(paste("KS, p = ", round(VAR.ks[[i]]$p.value,3)),
          side=1, line=2, cex=0.7)

    hist(norm.pit[,i], main="", breaks=seq(0,1, by=0.05),
         prob=TRUE, xlab="", ylab="")
    if(i==1) mtext("Normal", side=2, line=3)
    mtext(paste("KS, p = ", round(norm.ks[[i]]$p.value,3)),
          side=1, line=2, cex=0.7)
}

dev.off()


# Now compute the VRHs
library(ensembleBMA)


par(mfrow=c(3,2))
# Fix this....
MS1.vrh <- verifRankHist(t(MS[,,1]), as.vector(Y.future[,1]))
MS2.vrh <- verifRankHist(t(MS[,,2]), as.vector(Y.future[,2]))
VAR1.vrh <- verifRankHist(t(VAR[,,1]), Y.future[,1])
VAR2.vrh <- verifRankHist(t(VAR[,,2]), Y.future[,2])
norm1.vrh <- verifRankHist(t(normonly[,,1]), Y.future[,1])
norm2.vrh <- verifRankHist(t(normonly[,,2]), Y.future[,2])



# Make a custom plot so we can add our own labels
pdf(file="toyMSVRH.pdf", width=6, height=8)

par(mfrow=c(3,2), mai=c(0.4,0.25,0.25,0.25), omi=c(0,0.75,0.5,0))

hist(MS1.vrh, breaks=0:5001, prob=TRUE, main="", xlab="")
mtext("MSBVAR", side=2, line=3)
mtext("Series 1", side=3, line=1)
tmp <- ks.test(MS1.vrh, y="punif", min=0, max=5000)
mtext(paste("KS, p = ", round(tmp$p.value,3)),
      side=1, line=2, cex=0.7)

hist(MS2.vrh, breaks=0:5001, prob=TRUE, main="", xlab="")
mtext("Series 2", side=3, line=1)
tmp <- ks.test(MS2.vrh, y="punif", min=0, max=5000)
mtext(paste("KS, p = ", round(tmp$p.value,3)),
      side=1, line=2, cex=0.7)

hist(VAR1.vrh, breaks=0:5001, prob=TRUE, main="", xlab="")
mtext("VAR", side=2, line=3)
tmp <- ks.test(VAR1.vrh, y="punif", min=0, max=5000)
mtext(paste("KS, p = ", round(tmp$p.value,3)),
      side=1, line=2, cex=0.7)

hist(VAR2.vrh, breaks=0:5001, prob=TRUE, main="", xlab="")
tmp <- ks.test(VAR2.vrh, y="punif", min=0, max=5000)
mtext(paste("KS, p = ", round(tmp$p.value,3)),
      side=1, line=2, cex=0.7)

hist(norm1.vrh, breaks=0:5001, prob=TRUE, main="", xlab="")
mtext("Normal", side=2, line=3)
tmp <- ks.test(norm1.vrh, y="punif", min=0, max=5000)
mtext(paste("KS, p = ", round(tmp$p.value,3)),
      side=1, line=2, cex=0.7)

hist(norm2.vrh, breaks=0:5001, prob=TRUE, main="", xlab="")
tmp <- ks.test(norm2.vrh, y="punif", min=0, max=5000)
mtext(paste("KS, p = ", round(tmp$p.value,3)),
      side=1, line=2, cex=0.7)

dev.off()



########################################################
# Plot data and forecasts
########################################################
# Coverage intervals are 90%
VARse <- apply(VAR, c(2,3), quantile, probs=c(0.05, 0.5, 0.95))

normci <- apply(normonly, c(2,3), quantile, probs=c(0.05, 0.5, 0.95))
mixci <- apply(forc.norm, c(2,3), quantile, probs=c(0.05, 0.5, 0.95))

# Old version
## pdf(file="predictions90.pdf", width=6, height=5)

## par(mfrow=c(1,2), xpd=TRUE, mai=c(0.75, 0.65, 1, 0.15), cex=0.8)
## plot(ts(cbind(rd[,1],t(fout[,,1]),t(VARse[,,1]),t(normci[,,1]))),
##      plot.type="s", col=c(1,2,2,2,5,5,5,4,4,4),
##      lty=c(1,2,1,2,2,1,2,2,1,2), ylab="Series 1")

## lambda <- 0.005
## legend(par("usr")[1],
##        (1 + lambda) * par("usr")[4] - lambda * par("usr")[3],
##        c("Data", "MS-BVAR", "Frequentist", "Bivariate Normal"),
##        xjust = 0, yjust = 0, cex=0.8,
##        lwd=3, lty=1, col=c(par('fg'), 'red', 'cyan', 'blue'))

## plot(ts(cbind(rd[,2],t(fout[,,2]),t(VARse[,,2]),t(normci[,,2]))),
##      plot.type="s", col=c(1,2,2,2,5,5,5,4,4,4),
##      lty=c(1,2,1,2,2,1,2,2,1,2), ylab="Series 2")

## dev.off()

# New version
pdf(file="predictions90.pdf", width=8, height=5)

par(mfrow=c(1,3), xpd=TRUE, mai=c(0.75, 0.5, 0.75, 0.1), cex=0.8)
plot(ts(cbind(rd[,1],t(fout[,,1]),t(VARse[,,1]),t(normci[,,1]))),
     plot.type="s", col=c(1,2,2,2,5,5,5,4,4,4),
     lty=c(1,2,1,2,2,1,2,2,1,2), ylab="", main="Series 1", lwd=1.5)

plot(ts(cbind(rd[,2],t(fout[,,2]),t(VARse[,,2]),t(normci[,,2]))),
     plot.type="s", col=c(1,2,2,2,5,5,5,4,4,4),
     lty=c(1,2,1,2,2,1,2,2,1,2), ylab="", main="Series 2", lwd=1.5)

par(mar=rep(0,4))
plot.new()
legend("center",
       legend=c("Data", "MS-BVAR median forecast",
       "MS-BVAR 90% CI",
       "Frequentist median forecast",
       "Frequentist 90% CI",
       "Bivariate Normal median forecast",
       "Bivariate Normal 90% CI"),
       col= c(par('fg'), rep(c('red', 'cyan', 'blue'), each=2)),
       lty=c(1, rep(c(1,2), 3)), lwd=1.5)

dev.off()


# Save the results

save.image(file="toyMS.RData")


q(save="no")
